/*
 * wdx-jpg-comment
 * TotalCommander content plugin for JPG comment, EXIF and IPTC metadata
 *
 * Copyright  2006-2008 by Udo Liess, Thomas Beutlich
 * uses exiv2 (http://www.exiv2.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * http://www.fsf.org/licensing/licenses/gpl.txt
 */

#include "stdafx.h"
#include "contplug.h"
#include "resource.h"
#include "..\exiv2\src\exiv2.hpp"
#include "..\exiv2\src\image.hpp"

#define _detectstring "EXT=\"JPG\" | EXT=\"JPEG\" | EXT=\"TIF\" | EXT=\"TIFF\""
#define fieldcount 36

char* fieldnames[fieldcount] =
{
	"JPG Comment Complete",
	"JPG Comment Line",
	"-",
	"Artist",
	"Copyright 1",
	"Document Name",
	"Image Description",
	"Software",
	"User Comment",
	"-",
	"Creation Date",
	"Creation Time",
	"Caption",
	"Caption Writer",
	"Copyright 2",
	"Headline",
	"Special Instructions",
	"Subject",
	"Keywords",
	"Category",
	"Supplemental Categories",
	"Urgency",
	"Byline",
	"Byline Title",
	"Credits",
	"Source",
	"Contact",
	"Edit Status",
	"Fixture Id",
	"Object Name",
	"Location Name",
	"City",
	"Sublocation",
	"Province/State",
	"Country Name",
	"Transmission Reference"
};

int fieldtypes[fieldcount] =
{
	ft_string,
	ft_string,
	ft_boolean,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_boolean,
	ft_date,
	ft_time,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string,
	ft_string
};

char* keynames[fieldcount] =
{
	"Dummy", // dummy
	"Dummy", // dummy
	"Separator", // dummy
	"Exif.Image.Artist",
	"Exif.Image.Copyright",
	"Exif.Image.DocumentName",
	"Exif.Image.ImageDescription",
	"Exif.Image.Software",
	"Exif.Photo.UserComment",
	"Separator", // dummy
	"Iptc.Application2.DateCreated",
	"Iptc.Application2.TimeCreated",
	"Iptc.Application2.Caption",
	"Iptc.Application2.Writer",
	"Iptc.Application2.Copyright",
	"Iptc.Application2.Headline",
	"Iptc.Application2.SpecialInstructions",
	"Iptc.Application2.Subject",
	"Iptc.Application2.Keywords",
	"Iptc.Application2.Category",
	"Iptc.Application2.SuppCategory",
	"Iptc.Application2.Urgency",
	"Iptc.Application2.Byline",
	"Iptc.Application2.BylineTitle",
	"Iptc.Application2.Credit",
	"Iptc.Application2.Source",
	"Iptc.Application2.Contact",
	"Iptc.Application2.EditStatus",
	"Iptc.Application2.FixtureId",
	"Iptc.Application2.ObjectName",
	"Iptc.Application2.LocationName",
	"Iptc.Application2.City",
	"Iptc.Application2.SubLocation",
	"Iptc.Application2.ProvinceState",
	"Iptc.Application2.CountryName",
	"Iptc.Application2.TransmissionReference"
};

char* fieldunits_and_multiplechoicestrings[fieldcount] =
{
	"",
	"Line 1|Line 2|Line 3|Line 4|Line 5|Line 6|Line 7|Line 8|Line 9|Line 10|Line 11|Line 12|Line 13|Line 14|Line 15|Line 16",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	"",
	""
};

const int SeparatorIdx1 = 2;
const int SeparatorIdx2 = 9;
const int RepeatableIdx1 = 17;
const int RepeatableIdx2 = 18;
const int RepeatableIdx3 = 20;

typedef struct
{
	char* FieldValue;
	int FieldIndex;
	int maxlen;
	int flags;
	char* langidentifier;
} WndVar;

HINSTANCE DllInstance;
CRITICAL_SECTION criti;
BOOL loaded = FALSE; // criti initialized
volatile BOOL GetAborted;
char file_name2[MAX_PATH];
char* file_name = file_name2;
static std::string lngFileName = "";
Exiv2::Image::AutoPtr image;

BOOL CALLBACK DlgProc(HWND, UINT, WPARAM, LPARAM);

char* strlcpy(char* p, const char* p2, int maxlen)
{
	if ((int)strlen(p2) >= maxlen - 1)
	{
		strncpy_s(p, maxlen, p2, maxlen - 1);
		p[maxlen] = 0;
	}
	else
	{
		strcpy_s(p, maxlen, p2);
	}
	return p;
}

std::string Token(std::string Src, std::string Delim, int number)
{
	for (std::string::size_type First, Last = 0; First = Src.find_first_not_of(Delim, Last), First != std::string::npos ? Last = _cpp_min(Src.find_first_of(Delim, First), Src.size()), true : false;)
		if (!--number)
			return Src.substr(First, Last - First);
	return "";
}

BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID lpReserved)
{
	switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH :
			DllInstance = (HINSTANCE) hModule;
			DisableThreadLibraryCalls(DllInstance);
			char Path[_MAX_PATH];
			if (GetModuleFileName((HINSTANCE)hModule, Path, sizeof(Path)) != 0)
			{
				char d[_MAX_DRIVE];
				char p[_MAX_PATH];
				char f[_MAX_FNAME];
				_splitpath_s(Path, d, _MAX_DRIVE, p, _MAX_PATH, f, _MAX_FNAME, NULL, 0);
				lngFileName = std::string(d) + p + f + ".lng";
			}
			break;

		case DLL_PROCESS_DETACH :
			break;
	}
	return TRUE;
}

BOOL CALLBACK DlgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
		case WM_INITDIALOG :
			{
				WndVar* hWndVar = (WndVar*) lParam;
				SetWindowLongPtr(hWnd, GWL_USERDATA, (LONG)lParam);
				if (hWndVar && hWndVar->FieldValue && hWndVar->flags == editflags_initialize)
				{
					SetDlgItemText(hWnd, IDC_EDIT, hWndVar->FieldValue);
				}
				if (hWndVar && hWndVar->langidentifier)
				{
					char caption[MAX_PATH];
					GetPrivateProfileString(hWndVar->langidentifier, fieldnames[hWndVar->FieldIndex], fieldnames[hWndVar->FieldIndex], &caption[0], MAX_PATH, lngFileName.c_str());
					SetWindowText(hWnd, caption);
				}
				else
				{
					SetWindowText(hWnd, fieldnames[hWndVar->FieldIndex]);
				}
			}
			break;

		case WM_CLOSE :
			EndDialog(hWnd, IDCANCEL);
			break;

		case WM_COMMAND :
			if (wParam == IDOK)
			{
				WndVar * hWndVar = (WndVar*)(LONG_PTR)GetWindowLongPtr(hWnd, GWL_USERDATA);
				if (hWndVar && hWndVar->FieldValue)
				{
					GetDlgItemText(hWnd, IDC_EDIT, hWndVar->FieldValue, hWndVar->maxlen);
				}
				EndDialog(hWnd, IDOK);
			}
			else if (wParam == IDCANCEL)
			{
				EndDialog(hWnd, IDCANCEL);
			}
			break;

		default:
			return FALSE;
	}

	return TRUE;
}

int __stdcall ContentGetSupportedField(int FieldIndex, char* FieldName, char* Units, int maxlen)
{
	if (FieldIndex < 0 || FieldIndex >= fieldcount)
	{
		return ft_nomorefields;
	}
	strlcpy(FieldName, fieldnames[FieldIndex], maxlen - 1);
	strlcpy(Units, fieldunits_and_multiplechoicestrings[FieldIndex], maxlen - 1);
	return fieldtypes[FieldIndex];
}

int __stdcall ContentGetSupportedFieldFlags(int FieldIndex)
{
	switch (FieldIndex)
	{
		case -1:
		case 0:  // "JPG Comment"
		case 3:
		case 4:
		case 5:
		case 6:
		case 7:
		case 8:
		case 12:
		case 13:
		case 14:
		case 15:
		case 16:
		case 17:
		case 18:
		case 19:
		case 20:
		case 21:
		case 22:
		case 23:
		case 24:
		case 25:
		case 26:
		case 27:
		case 28:
		case 29:
		case 30:
		case 31:
		case 32:
		case 33:
		case 34:
		case 35:
			return contflags_edit | contflags_fieldedit;
			break;

		case 10:
		case 11:
			return contflags_edit;
			break;

		default:
			return contflags_substsize;
	}
}

int __stdcall ContentGetDetectString(char* DetectString, int maxlen)
{
	strlcpy(DetectString, _detectstring, maxlen);
	return 0;
}

int __stdcall ContentGetValue(char* FileName, int FieldIndex, int UnitIndex, void* FieldValue, int maxlen, int flags)
{
	if (flags & CONTENT_DELAYIFSLOW)
	{
		return ft_delayed;
	}

	if (!loaded)
	{
		InitializeCriticalSection(&criti);
		EnterCriticalSection(&criti);
		loaded = TRUE;
		LeaveCriticalSection(&criti);
	}

	if (strcmp(file_name, FileName))
	{
		EnterCriticalSection(&criti);
		strcpy(file_name, FileName);
		GetAborted = FALSE;
		LeaveCriticalSection(&criti);
	}

	if (FieldIndex < SeparatorIdx1) // "JPG Comment"
	{
		if (GetAborted)
		{
			return ft_fieldempty;
		}
		try
		{
			image = Exiv2::ImageFactory::open(FileName);
			if (image->good())
			{
				image->readMetadata();
				std::string comment = image->comment();
				if (!FieldIndex) // JPG Comment Complete
				{
					strlcpy((char*)FieldValue, comment.c_str(), maxlen - 1);
				}
				else // JPG Comment Line
				{
					strlcpy((char*)FieldValue, Token(comment, "\n\r", UnitIndex + 1).c_str(), maxlen - 1);
				}
			}
			else
			{
				return ft_fileerror;
			}
		}
		catch (...)
		{
			return ft_fileerror;
		}
	}
	else if (FieldIndex < fieldcount)
	{
		if (GetAborted)
		{
			return ft_fieldempty;
		}

		if (FieldIndex == SeparatorIdx1 || FieldIndex == SeparatorIdx2)
		{
			return ft_nosuchfield;
		}

		try
		{
			image = Exiv2::ImageFactory::open(FileName);
			if (image->good())
			{
				image->readMetadata();
				if (FieldIndex < SeparatorIdx2)
				{
					Exiv2::ExifData &exifData = image->exifData();
					Exiv2::ExifKey exifKey(keynames[FieldIndex]);
					Exiv2::ExifData::const_iterator iter = exifData.findKey(exifKey);
					if (iter != exifData.end())
					{
						switch (fieldtypes[FieldIndex])
						{
							case ft_string:
								strlcpy((char*)FieldValue, iter->value().toString().c_str(), maxlen - 1);
								break;

							default:
								return ft_fileerror;
						}
					}
					else
					{
						return ft_fieldempty;
					}
				}
				else
				{
					Exiv2::IptcData &iptcData = image->iptcData();
					Exiv2::IptcKey iptcKey(keynames[FieldIndex]);
					Exiv2::IptcData::iterator iter = iptcData.findKey(iptcKey);
					if (iter == iptcData.end())
					{
						return ft_fieldempty;
					}
					switch (fieldtypes[FieldIndex])
					{
						case ft_string:
							if (FieldIndex == RepeatableIdx1 || FieldIndex == RepeatableIdx2 || FieldIndex == RepeatableIdx3)
							{
								BOOL first = TRUE;
								for (iter = iter; iter != iptcData.end(); ++iter)
								{
									if (first)
									{
										first = FALSE;
										strlcpy((char*)FieldValue, iter->value().toString().c_str(), maxlen - 1);
									}
									else
									{
										if (iter->key() != std::string(keynames[FieldIndex]))
										{
											continue;
										}

										strcat((char*)FieldValue, "\r\n");
										strcat((char*)FieldValue, iter->value().toString().c_str());
									}
								}
							}
							else
							{
								strlcpy((char*)FieldValue, iter->value().toString().c_str(), maxlen - 1);
							}
							break;

						case ft_date:
						{
							int Year = 0 , Month = 0, Day = 0;
							sscanf(iter->value().toString().c_str(), "%u-%u-%u", &Year, &Month, &Day);
							((pdateformat)FieldValue)->wYear = Year;
							((pdateformat)FieldValue)->wMonth = Month;
							((pdateformat)FieldValue)->wDay = Day;
							break;
						}

						case ft_time:
						{
							int Hour = 0 , Minute = 0, Second = 0, HourOffset = 0 , MinOffset = 0;
							char Offset = '+';
							sscanf(iter->value().toString().c_str(), "%u:%u:%u%c%u:%u", &Hour, &Minute, &Second, &Offset, &HourOffset, &MinOffset);
							((ptimeformat)FieldValue)->wHour = Hour;
							((ptimeformat)FieldValue)->wMinute = Minute;
							((ptimeformat)FieldValue)->wSecond = Second;
							break;
						}

						default:
							return ft_fileerror;
					}
				}
			}
			else
			{
				return ft_fileerror;
			}
		}
		catch (...)
		{
			return ft_fileerror;
		}
	}
	else
	{
		return ft_nosuchfield;
	}

	return fieldtypes[FieldIndex];  // very important!
}

int __stdcall ContentEditValue(HWND ParentWin, int FieldIndex, int UnitIndex, int FieldType, void* FieldValue, int maxlen, int flags, char* langidentifier)
{
	int retVal = ft_nosuchfield;

	if (FieldValue && FieldIndex >= 0 && FieldIndex < fieldcount && FieldIndex != SeparatorIdx1 && FieldIndex != SeparatorIdx2 && FieldIndex != 10 && FieldIndex != 11)
	{
		WndVar * hWndVar = (WndVar *) malloc(sizeof(WndVar));
		if (hWndVar)
		{
			hWndVar->FieldValue = (char*) FieldValue;
			hWndVar->FieldIndex = FieldIndex;
			hWndVar->maxlen = maxlen;
			hWndVar->flags = flags;
			hWndVar->langidentifier = langidentifier;
			INT_PTR result = DialogBoxParam(DllInstance, MAKEINTRESOURCE(IDD_EDIT), ParentWin, DlgProc, (LPARAM) hWndVar);
			free(hWndVar);
			retVal = result == IDOK ? ft_setsuccess : ft_setcancel;
		}
	}

	return retVal;
}

void __stdcall ContentPluginUnloading(void)
{
	if (loaded)
	{
		EnterCriticalSection(&criti);
		loaded = FALSE;
		LeaveCriticalSection(&criti);
		DeleteCriticalSection(&criti);
	}
}

void __stdcall ContentSetDefaultParams(ContentDefaultParamStruct* dps)
{
}

int __stdcall ContentSetValue(char* FileName, int FieldIndex, int UnitIndex, int FieldType, void* FieldValue, int flags)
{
	if ((FileName) && (FieldIndex < fieldcount))
	{
		try
		{
			if (flags & setflags_first_attribute)
			{
				image = Exiv2::ImageFactory::open(FileName);
				if (image->good())
				{
					image->readMetadata();
				}
				else
				{
					return ft_fileerror;
				}
			}

			switch (FieldIndex)
			{
				/* JPG Comment */
				case 0:
					switch (FieldType)
					{
						case ft_string:
							image->setComment(std::string((char *) FieldValue));
							break;

						default:
							return ft_nosuchfield;
					}
					break;

				/* EXIF */
				case 3:
				case 4:
				case 5:
				case 6:
				case 7:
				case 8:
					switch (FieldType)
					{
						case ft_string:
						{
							Exiv2::ExifData &exifData = image->exifData();
							exifData[keynames[FieldIndex]] = std::string((char *) FieldValue);
							break;
						}

						default:
							return ft_nosuchfield;
					}
					break;

				/* IPTC */
				case 10:
				case 11:
				case 12:
				case 13:
				case 14:
				case 15:
				case 16:
				case 17:
				case 18:
				case 19:
				case 20:
				case 21:
				case 22:
				case 23:
				case 24:
				case 25:
				case 26:
				case 27:
				case 28:
				case 29:
				case 30:
				case 31:
				case 32:
				case 33:
				case 34:
				case 35:
					switch (FieldType)
					{
						case ft_string:
						{
							Exiv2::IptcData &iptcData = image->iptcData();
							if (FieldIndex == RepeatableIdx1 || FieldIndex == RepeatableIdx2 || FieldIndex == RepeatableIdx3)
							{
								Exiv2::IptcKey iptcKey(keynames[FieldIndex]);
								while (TRUE)
								{
									Exiv2::IptcData::iterator iter = iptcData.findKey(iptcKey);
									if (iter == iptcData.end())
									{
										break;
									}
									else
									{
										iptcData.erase(iter);
									}
								}

								int Idx = 1;
								std::string Data = Token(std::string((char *) FieldValue), "\n\r", Idx);
								Exiv2::Value::AutoPtr value = Exiv2::Value::create(Exiv2::TypeInfo::typeId("String"));

								while (Data != "")
								{
									value->read(Data);
									iptcData.add(iptcKey, value.get());
									Data = Token(std::string((char *) FieldValue), "\n\r", ++Idx);
								}
							}
							else
							{
								iptcData[keynames[FieldIndex]] = std::string((char *) FieldValue);
							}
							break;
						}

						case ft_date:
						{
							Exiv2::IptcData &iptcData = image->iptcData();
							char dateStr[11] = "\0";
							wsprintf(dateStr, "%4u-%02u-%02u", ((pdateformat)FieldValue)->wYear, ((pdateformat)FieldValue)->wMonth, ((pdateformat)FieldValue)->wDay);
							iptcData[keynames[FieldIndex]] = std::string(dateStr);
							break;
						}

						case ft_time:
						{
							Exiv2::IptcData &iptcData = image->iptcData();
							char timeStr[15] = "\0";
							TIME_ZONE_INFORMATION tzi = {0};
							GetTimeZoneInformation(&tzi);
							BOOL Ahead = tzi.Bias < 0;
							int HourOffset = Ahead ? -tzi.Bias/60 : tzi.Bias/60;
							int MinOffset = Ahead ? -tzi.Bias%60 : tzi.Bias%60;
							wsprintf(timeStr, "%02u:%02u:%02u%c%02u:%02u", ((ptimeformat)FieldValue)->wHour, ((ptimeformat)FieldValue)->wMinute, ((ptimeformat)FieldValue)->wSecond, Ahead ? '+' : '-', HourOffset, MinOffset);
							iptcData[keynames[FieldIndex]] = std::string(timeStr);
							break;
						}

						default:
							return ft_nosuchfield;

					}
					break;

				default:
					return ft_nosuchfield;
			}

			if (flags & setflags_last_attribute)
			{
				FILETIME ftCreate, ftAccess, ftWrite;
				HANDLE fh = CreateFile(FileName, GENERIC_ALL, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, 0);
				if (fh != INVALID_HANDLE_VALUE)
				{
					BOOL result = GetFileTime(fh, &ftCreate, &ftAccess, &ftWrite);
					CloseHandle(fh);
					if (!result)
					{
						return ft_fileerror;
					}
				}

				image->writeMetadata();

				fh = CreateFile(FileName, GENERIC_ALL, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, 0);
				if (fh != INVALID_HANDLE_VALUE)
				{
					BOOL result = SetFileTime(fh, (LPFILETIME) NULL, (LPFILETIME) NULL, &ftWrite);
					CloseHandle(fh);
					return result ? ft_setsuccess : ft_fileerror;
				}
				else
				{
					return ft_fileerror;
				}
			}
			else
			{
				return ft_setsuccess;
			}
		}
		catch (...)
		{
			return ft_fileerror;
		}
	}
	else
	{
		return ft_nosuchfield;
	}
}

void __stdcall ContentStopGetValue(char* FileName)
{
	EnterCriticalSection(&criti);
	GetAborted = TRUE;
	LeaveCriticalSection(&criti);
}

int __stdcall ContentGetDefaultSortOrder(int FieldIndex)
{
	return ft_notsupported;
}

void __stdcall ContentSendStateInformation(int state, char* path)
{
}
